/*
 * Universal Smart Keyboard driver for Linux
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/slab.h>
#include <linux/module.h>
#include <linux/input.h>
#include <linux/init.h>
#include <linux/serio.h>

#define DRIVER_DESC	"Universal Smart Keyboard driver"

MODULE_AUTHOR("Osamu KURATI");
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");

#define N_KEYCODE 128

static unsigned char uskbd_keycode[N_KEYCODE] = {
    /* 00 */ 0, KEY_1, KEY_2, KEY_3, KEY_4, KEY_5, KEY_6, KEY_Q,
    /* 08 */ KEY_W, KEY_E, KEY_R, KEY_T, KEY_A, KEY_S, KEY_D, KEY_F,
    /* 10 */ KEY_G, KEY_Z, KEY_X, KEY_C, KEY_V, KEY_B, KEY_SPACE, KEY_TAB,
    /* 18 */ KEY_LEFTCTRL, KEY_GRAVE, KEY_7, KEY_8, KEY_9, KEY_0, KEY_MINUS, KEY_EQUAL,
    /* 20 */ KEY_BACKSPACE, KEY_Y, KEY_U, KEY_I, KEY_O, KEY_P, KEY_LEFTBRACE, KEY_RIGHTBRACE,
    /* 28 */ KEY_H, KEY_J, KEY_K, KEY_L, KEY_SEMICOLON, KEY_APOSTROPHE, KEY_ENTER, KEY_N,
    /* 30 */ KEY_M, KEY_COMMA, KEY_DOT, KEY_SLASH, KEY_UP, KEY_SPACE, KEY_BACKSLASH, KEY_LEFT,
    /* 38 */ KEY_DOWN, KEY_RIGHT, KEY_LEFTCTRL, KEY_RIGHTSHIFT, KEY_LEFTSHIFT, KEY_DELETE, 0, KEY_LEFTALT,
    /* F00 */ 0, KEY_F1, KEY_F2, KEY_F3, KEY_F4, KEY_F5, KEY_F6, KEY_Q,
    /* F08 */ KEY_W, KEY_E, KEY_R, KEY_T, KEY_A, KEY_S, KEY_D, KEY_F,
    /* F10 */ KEY_G, KEY_Z, KEY_X, KEY_C, KEY_V, KEY_B, KEY_SPACE, KEY_TAB,
    /* F18 */ KEY_LEFTCTRL, KEY_GRAVE, KEY_F7, KEY_F8, KEY_F9, KEY_F10, KEY_F11, KEY_F12,
    /* F20 */ KEY_BACKSPACE, KEY_Y, KEY_U, KEY_I, KEY_O, KEY_P, KEY_LEFTBRACE, KEY_RIGHTBRACE,
    /* F28 */ KEY_H, KEY_J, KEY_K, KEY_L, KEY_SEMICOLON, KEY_APOSTROPHE, KEY_ENTER, KEY_N,
    /* F30 */ KEY_M, KEY_COMMA, KEY_DOT, KEY_SLASH, KEY_PAGEUP, KEY_SPACE, KEY_BACKSLASH, KEY_HOME,
    /* F38 */ KEY_PAGEDOWN, KEY_END, KEY_LEFTCTRL, KEY_RIGHTSHIFT, KEY_LEFTSHIFT, KEY_DELETE, 0, KEY_LEFTALT,
};

static char *uskbd_name = "Universal Smart Keyboard";

struct uskbd {
    unsigned char keycode[N_KEYCODE];
    struct input_dev dev;
    struct serio *serio;
    char phys[32];
    int fn_state;
};

static irqreturn_t uskbd_interrupt(struct serio *serio,
		unsigned char data, unsigned int flags, struct pt_regs *regs)
{
	struct uskbd *uskbd = serio_get_drvdata(serio);
	int pressed;
	unsigned char *keycodes;

	if (data < 0x40) {
	    pressed = 1;
	} else {
	    pressed = 0;
	    data -= 0x3f;
	}

	if (data == 0x3e) {
	    /* FN */
	    uskbd->fn_state = pressed;
	} else {
	    keycodes = uskbd->keycode;
	    if (uskbd->fn_state) {
		keycodes += 0x40;
	    }

	    input_regs(&uskbd->dev, regs);
	    input_report_key(&uskbd->dev, keycodes[data], pressed);
	    input_sync(&uskbd->dev);
	}

	return IRQ_HANDLED;

}

static int uskbd_connect(struct serio *serio, struct serio_driver *drv)
{
	struct uskbd *uskbd;
	int i;
	int err;

	if (!(uskbd = kmalloc(sizeof(struct uskbd), GFP_KERNEL)))
		return -ENOMEM;

	memset(uskbd, 0, sizeof(struct uskbd));

	uskbd->dev.evbit[0] = BIT(EV_KEY) | BIT(EV_REP);

	uskbd->serio = serio;
	uskbd->fn_state = 0;

	init_input_dev(&uskbd->dev);
	uskbd->dev.keycode = uskbd->keycode;
	uskbd->dev.keycodesize = sizeof(unsigned char);
	uskbd->dev.keycodemax = ARRAY_SIZE(uskbd_keycode);
	uskbd->dev.private = uskbd;

	serio_set_drvdata(serio, uskbd);

	err = serio_open(serio, drv);
	if (err) {
		serio_set_drvdata(serio, NULL);
		kfree(uskbd);
		return err;
	}

	memcpy(uskbd->keycode, uskbd_keycode, sizeof(uskbd->keycode));
	for (i = 0; i < N_KEYCODE; i++)
		set_bit(uskbd->keycode[i], uskbd->dev.keybit);
	clear_bit(0, uskbd->dev.keybit);

	sprintf(uskbd->phys, "%s/input0", serio->phys);

	uskbd->dev.name = uskbd_name;
	uskbd->dev.phys = uskbd->phys;
	uskbd->dev.id.bustype = BUS_RS232;
	uskbd->dev.id.vendor = SERIO_UNKNOWN;
	uskbd->dev.id.product = 0x0001;
	uskbd->dev.id.version = 0x0001;
	uskbd->dev.dev = &serio->dev;

	input_register_device(&uskbd->dev);

	printk(KERN_INFO "input: %s on %s\n", uskbd_name, serio->phys);

	return 0;
}

static void uskbd_disconnect(struct serio *serio)
{
	struct uskbd *uskbd = serio_get_drvdata(serio);

	input_unregister_device(&uskbd->dev);
	serio_close(serio);
	serio_set_drvdata(serio, NULL);
	kfree(uskbd);
}

static struct serio_device_id uskbd_serio_ids[] = {
	{
		.type	= SERIO_RS232,
		.proto	= SERIO_UNKNOWN,
		.id	= SERIO_ANY,
		.extra	= SERIO_ANY,
	},
	{ 0 }
};

MODULE_DEVICE_TABLE(serio, uskbd_serio_ids);

static struct serio_driver uskbd_drv = {
	.driver		= {
		.name	= "usmartkbd",
	},
	.description	= DRIVER_DESC,
	.id_table	= uskbd_serio_ids,
	.interrupt	= uskbd_interrupt,
	.connect	= uskbd_connect,
	.disconnect	= uskbd_disconnect,
};

static int __init uskbd_init(void)
{
	serio_register_driver(&uskbd_drv);
	return 0;
}

static void __exit uskbd_exit(void)
{
	serio_unregister_driver(&uskbd_drv);
}

module_init(uskbd_init);
module_exit(uskbd_exit);
